package lightcontainer.domains.client;

import lightcontainer.interfaces.ClientHandlerFFE;
import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.interfaces.UnicastCHR;
import lightcontainer.protocol.ProtocolReader;
import lightcontainer.protocol.ProtocolWriter;
import lightcontainer.protocol.rules.reader.SigninRule;
import lightcontainer.protocol.rules.reader.SignoutRule;
import lightcontainer.protocol.rules.reader.SignupRule;
import lightcontainer.protocol.rules.writer.SignErrorRule;
import lightcontainer.utils.Log;

import java.io.IOException;
import java.net.Socket;

/**
 * ClientHandler
 * <p>
 * <!> UNICAST CLIENT <!>
 * Class communicating with the client, and
 * intercepting and sending files to the client.
 *
 * @author Jérémi NIHART <j.nihart@student.helmo.be>
 * @version 1.1
 * @see Runnable
 * @see AutoCloseable
 * @since 1.0
 */
public class ClientHandler extends UnicastThread implements AutoCloseable {
    // Variables
    private final ClientHandlerFFE fileFrontEnd;
    private final ProtocolRepository protocolRep;
    private final UnicastCHR repository;
    private ProtocolWriter.ProtocolResult response;

    // Constructor
    public ClientHandler(UnicastCHR repository, Socket client, ClientHandlerFFE ffe, ProtocolRepository protocolRep, Context context) {
        super(client, context);
        this.repository = repository;
        this.fileFrontEnd = ffe;
        this.protocolRep = protocolRep;
    }

    /**
     * Thread Function
     * Start the dialogue with the client.
     *
     * @since 1.0
     */
    @Override
    public void run() {
        // Set process to true
        this.setRunning(true);
        // Process while
        while (this.isRunning()) {
            // Signifie le démarrage d'une nouvelle rquête
            getContext().newBundle();
            try {
                String command = this.readLine();
                if (command != null) {
                    Log.getInstance().infoln("Client: " + command);
                } else {
                    this.repository.disconnect(this);
                    break;
                }

                ProtocolReader.ProtocolResult ruleResult = this.protocolRep.executeReader(getContext(), command + "\r\n");
                if (ruleResult == null) {
                    this.repository.disconnect(this);
                    break;
                }

                if (checkAccess(ruleResult)) {
                    // Lecture du fichier client
                    ruleResult.read(this.getInputStream());
                    ProtocolWriter.ProtocolResult writerCommand = ruleResult.getResultCommand();
                    // TODO : Vérifier que le StorBackEnd demandé (Pas toujours demandé) est disponible
                    if (ruleResult.getReceiver() == ProtocolReader.ResultCmdReceiver.STOREBACKEND && !fileFrontEnd.canExecuteCommand(ruleResult.getRequestDomain())) {
                        this.print(ruleResult.onNotExecutable(getContext())); // Renvoie au client
                    } else if (ruleResult.getReceiver() == ProtocolReader.ResultCmdReceiver.STOREBACKEND) {
                        this.fileFrontEnd.newCommand(getContext(), writerCommand, ruleResult.getRequestDomain()); // Envoie dans la file de tâche FileFrontEnd en attente d'un traitement d'un StorBackEnd
                        // Attend la fin de la réalisation de la tâche
                        waitTaskResponse();
                        if (this.response != null) {
                            this.write(this.response.getCommand()); // Renvoie au client
                            this.response.write(this.getOutputStream()); // Ecrit au client si nécessaire
                        } else {
                            this.print(ruleResult.onNotExecutable(getContext())); // Renvoie au client
                        }
                    } else {
                        this.print(writerCommand.getCommand()); // Renvoie au client
                    }
                } else {
                    accessDenied();
                }
            } catch (IOException e) {
                this.close();
            }
        }
    }

    /**
     * Permet de vérifier si le client possède l'accès demandé
     *
     * @param ruleResult La règle
     * @return TRUE si le client possède l'accès demandé
     */
    private boolean checkAccess(ProtocolReader.ProtocolResult ruleResult) {
        checkSignout(ruleResult);
        if (getContext().isConnected())
            return true;
        try {
            ruleResult
                    .getClass()
                    .asSubclass(SignupRule.Result.class);
            return true;
        } catch (ClassCastException e) {
            try {
                ruleResult.getClass().asSubclass(SigninRule.Result.class);
                return true;
            } catch (ClassCastException ignored) { }
        }
        return false;
    }

    /**
     * Lorsque l'accès à été refusé.
     */
    private void accessDenied() {
        ProtocolWriter.ProtocolResult signErrorResult = this.protocolRep.executeWriter(getContext(), SignErrorRule.NAME);
        this.write(signErrorResult.getCommand()); // Envoie SignError car echec de la connection
    }

    /**
     * Vérifie s'il s'âgit d'une demande de déconnexion
     *
     * @param ruleResult
     */
    private void checkSignout(ProtocolReader.ProtocolResult ruleResult) {
        try {
            ruleResult.getClass().asSubclass(SignoutRule.Result.class);
            this.repository.disconnect(this);
        } catch (ClassCastException ignored) {}
    }

    /**
     * Permet au Client d'attendre la fin de la réalisation de sa tâche
     */
    private void waitTaskResponse() {
        synchronized (this) {
            try {
                this.wait();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Permet d'envoyer la réponse au client.
     *
     * @param response La réponse
     */
    public void respond(ProtocolWriter.ProtocolResult response) {
        this.response = response;
        synchronized (this) {
            this.notify();
        }
    }

    /**
     * AutoClosable Function
     * Close the Client thread and resources.
     *
     * @since 1.0
     */
    @Override
    public void close() {
        if (this.isRunning()) {
            try {
                super.close();
                this.repository.disconnect(this);
                Log.getInstance().infof("[CLIENT] %s s'est déconnecté\n", getContext().getLogin());
            } catch (IOException e) {
                Log.getInstance().infoln("[CH] Error while closing client.");
            }
        }
    }

    public String getLogin() {
        return this.getContext().getLogin();
    }

}
