package lightcontainer.domains.client;

import lightcontainer.repository.ReadOnlyClientRepository;
import lightcontainer.storage.File;
import lightcontainer.storage.ReadOnlyFile;
import lightcontainer.storage.Repository;
import lightcontainer.utils.AES_GCM;
import lightcontainer.utils.BCryptHasher;
import lightcontainer.utils.ShaHasher;

import java.util.List;
import java.util.Set;

/**
 * Contexte associé à la requête d'un utilisateur.
 * Permet de récupérer les données à celui-ci et d'effectuer des actions sur le context courant.
 */
public class Context {
    // Variables
    private final Repository repository;
    private RequestBundle requestBundle;
    private ReadOnlyClientRepository clientRepository;

    /**
     * Login de l'utilisateur
     */
    private String login;

    /**
     * Domain s'occupant de la requête
     */
    private String domain;

    // Constructeur
    public Context(Repository repository, ReadOnlyClientRepository clientRepository) {
        this.repository = repository;
        this.clientRepository = clientRepository;
    }


    /**
     * Permet de créer un utilisateur.
     *
     * @param login    Login de l'utilisateur
     * @param password Mot de passe de l'utilisateur
     * @return TRUE si l'utilisateur a pu être créé
     */
    public boolean createUser(String login, String password) {
        if (clientRepository.hasClient(login)) { // Empêcher double connection
            return false;
        }

        try {
            String key = AES_GCM.generateSecretKey();
            String hashedPassword = BCryptHasher.hashPassword(password);
            if (this.repository.addUser(login, hashedPassword, key)) {
                this.login = login;
                return true;
            }
        } catch (AES_GCM.AesGcmException ignored) {
        }
        return false;
    }

    /**
     * Login de l'utilisateur
     *
     * @return Login de l'utilisateur
     */
    public String getLogin() {
        return this.login;
    }


    /**
     * Permet de demander la connexion de l'utilisateur
     *
     * @param login    Login
     * @param password Mot de passe
     * @return TRUE si l'utilisateur a été authentifié
     */
    public boolean signIn(String login, String password) {
        if (login == null || password == null) {
            return false;
        }

        // Empêcher double connection
        if (!clientRepository.hasClient(login) && this.repository.verifyUser(login, password)) {
            this.login = login;
            return true;
        }
        return false;
    }

    /**
     * Permet de savoir si l'utilisateur s'est connecté (Avec login et mdp)
     *
     * @return TRUE si l'utilisateur est connecté
     */
    public boolean isConnected() {
        return this.getLogin() != null;
    }


    /**
     * Clé AES de l'utilisateur
     *
     * @return Clé AES de l'utilisateur
     */
    public String getAesKey() {
        return this.repository.getUserAesKey(getLogin());
    }

    /**
     * Permet de régénérer le Requestbundle courant. Méthode appelée lorsque l'on souhaite
     */
    public void newBundle() {
        requestBundle = new RequestBundle();
    }

    /**
     * Permet d'ajouter des données pour la requête courante
     *
     * @param key   La clé permettant de retrouver la valeur
     * @param value La valeur associée à la clé
     */
    public void putDataString(String key, String value) {
        requestBundle.putString(key, value);
    }

    /**
     * Permet d'ajouter des données pour la requête courante
     *
     * @param key   La clé permettant de retrouver la valeur
     * @param value La valeur associée à la clé
     */
    public void putDataInt(String key, int value) {
        requestBundle.putInt(key, value);
    }

    /**
     * @param key   La clé permettant de retrouver la valeur
     * @param value La valeur associée à la clé
     */
    public void putDataLong(String key, long value) {
        requestBundle.putLong(key, value);
    }

    /**
     * Permet de récupérer des données pour la requête courante
     *
     * @param key La clé permettant de retrouver la valeur
     * @return La valeur associée à la clé ou null
     */
    public String getDataString(String key) {
        return requestBundle.getString(key);
    }

    /**
     * Permet de récupérer des données pour la requête courante
     *
     * @param key La clé permettant de retrouver la valeur
     * @return La valeur associée à la clé
     */
    public int getDataInt(String key) {
        return requestBundle.getInt(key);
    }

    /**
     * @param key La clé permettant de retrouver la valeur
     * @return La valeur associée à la clé
     */
    public long getDataLong(String key) {
        return requestBundle.getLong(key);
    }

    /**
     * Permet d'ajouter un fichier à l'utilisateur
     *
     * @param fileName     Nom du fichier hashé
     * @param fileNameSalt Salt appliqué sur le nom du fichier
     * @param size         Taille du fichier
     * @param iv           IV du fichier
     * @param domain       Domain dans lequel est stocké le fichier
     * @return TRUE si le fichier a pu être enregistré.
     */
    public boolean addFile(String fileName, String fileNameSalt, long size, String iv, String domain) {
        return this.repository.addFileFor(new File(fileName, fileNameSalt, size, iv, Set.of(domain)), getLogin());
    }

    public String getHashedFileName(String fileName) {
        ReadOnlyFile file = repository.getFileOf(fileName, login);
        if (file != null) {
            String completeFileName = login + "_" + fileName;
            ShaHasher hasher = new ShaHasher(completeFileName);
            return hasher.fromSalt(hasher.saltToByte(file.getFileNameSalt()));
        }
        return null;
    }

    public ReadOnlyFile getFile(String fileName) {
        return this.repository.getFileOf(fileName, login);
    }

    public List<String> getStringifiedFilesOf() {
        return repository.getStringifiedFilesOf(login);
    }

    public boolean canAddFile() {
        return repository.canAddFile(login);
    }

    public String getDomain() {
        return this.domain;
    }

    public void setDomain(String domain) {
        this.domain = domain;
    }

    public void deleteFile(String deletedFileName) {
        repository.deleteFileOf(deletedFileName, login);
    }
}
