package lightcontainer.domains.client;

import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.interfaces.StoreProcessorFFE;
import lightcontainer.protocol.ProtocolReader;
import lightcontainer.protocol.ProtocolWriter;
import lightcontainer.utils.Log;

import java.io.IOException;
import java.net.Socket;
import java.time.LocalDateTime;
import java.util.Objects;

/**
 * StoreProcessor
 * <p>
 * <!> MULTICAST CLIENT <!>
 * Class communicating with the storebackend and
 * processing the tasks in the FileFrontEnd
 *
 * @author Jérémi NIHART <j.nihart@student.helmo.be>
 * @version 1.1
 * @see Runnable
 * @see AutoCloseable
 * @since 1.0
 */
public class StoreProcessor extends UnicastThread implements AutoCloseable {
    // Variables
    private final StoreProcessorFFE fileFrontEnd;
    private final String domain;
    private LocalDateTime lastAnnounce;

    private ProtocolWriter.ProtocolResult protocolResult;
    private final ProtocolRepository protocolRep;

    // Constructor
    public StoreProcessor(Socket socket, String domain, StoreProcessorFFE ffe, ProtocolRepository protocolRep) {
        super(socket);
        this.domain = domain;
        this.fileFrontEnd = ffe;
        this.protocolRep = protocolRep;
    }

    /**
     * Thread Function
     * Start the dialogue with the storebackend.
     *
     * @since 1.0
     */
    @Override
    public void run() {
        // Set process to true
        this.setRunning(true);
        // Process while
        while (this.isRunning()) {
            try {
                if (protocolResult == null) { // Si on n'a pas encore la commande à envoyer
                    waitAction();
                }
                Log.getInstance().infoln("[SBE] Envoie commande : " + protocolResult.getCommand());

                // Request
                this.write(protocolResult.getCommand());

                try {
                    protocolResult.write(this.getOutputStream());
                } catch (IOException writeException) { // Si SBE fermé
                    // Envoie au client que la requête n'a pu être traitée
                    alertAvailable(null);
                    break;
                }
                // Response
                String responseCommand = this.readLine();
                if (responseCommand != null && !responseCommand.isBlank()) {
                    responseCommand += "\r\n";
                    ProtocolReader.ProtocolResult responseResult = protocolRep.executeReader(getContext(), responseCommand);
                    if (responseResult != null) {
                        Log.getInstance().infoln("StoreBackEnd (" + domain + ") response to client: " + responseResult.getResultCommand());
                        responseResult.read(
                                this.getInputStream()
                        );

                        alertAvailable(responseResult.getResultCommand());
                    } else {
                        Log.getInstance().infoln("StoreBackEnd result: Commande null");
                        alertAvailable(null);
                    }
                } else {
                    Log.getInstance().infoln("StoreBackEnd: Commande null");
                    alertAvailable(null);
                }

            } catch (IOException exception) {
                Log.getInstance().infoln("[ERROR] Problem with SBE (" + domain + ") : " + exception.getMessage());
                this.close();
            }
        }

        // Closing SBE
        try {
            this.close();
            this.fileFrontEnd.onStoreDisconnect(this.domain);
        } catch (Exception e) {
            Log.getInstance().infoln("[ERROR] Error while closing SBE (" + domain + ") : " + e.getMessage());
        }
    }

    /**
     * Permet de demander au StoreBackEnd d'effectuer une commande
     *
     * @param protocolResult La commande à effectuer
     */
    public void executeCommand(Context context, ProtocolWriter.ProtocolResult protocolResult) {
        synchronized (this) {
            this.protocolResult = protocolResult;
            setContext(context);
            this.notify();
        }
    }

    /**
     * Permet de déclarer le StoreBackEnd disponible
     */
    private void alertAvailable(ProtocolWriter.ProtocolResult responseCommand) {
        synchronized (this) {
            this.protocolResult = null;
            if (responseCommand == null) {
                this.setRunning(false);
            }
            fileFrontEnd.onStoreAvailable(this.getDomain(), responseCommand);
        }
    }

    /**
     * Permet au Store d'attendre une action à réaliser
     */
    private void waitAction() {
        synchronized (this) {
            try {
                this.wait();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * AutoClosable Function
     * Close the Storage thread and resources.
     *
     * @since 1.0
     */
    @Override
    public void close() {
        if (this.isRunning()) {
            this.setRunning(false);
            Log.getInstance().infoln("[SBE] Fermeture de " + domain);
        }
    }

    /**
     * Compare two StoreProcessor object to check if equals.
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        StoreProcessor that = (StoreProcessor) o;
        return Objects.equals(domain, that.domain);
    }

    /**
     * Get hash code
     */
    @Override
    public int hashCode() {
        return Objects.hash(domain);
    }

    public boolean canProcessTask() {
        return this.protocolResult == null; // Vérifier si tâche veut ce SBE
    }

    public String getDomain() {
        return this.domain;
    }

    /**
     * Permet de mettre à jours la dernière annonce de ce SBE au FFE
     */
    public void setLastAnnounce(LocalDateTime lastAnnounce) {
        this.lastAnnounce = lastAnnounce;
    }

    /**
     * Permet de mettre à jours la dernière annonce de ce SBE au FFE
     */
    public LocalDateTime getLastAnnounce() {
        return lastAnnounce;
    }
}
