package lightcontainer.domains.client;

import java.io.*;
import java.net.Socket;
import java.nio.charset.StandardCharsets;

public abstract class UnicastThread extends Thread implements AutoCloseable {
    // Variables
    private final Socket socket;
    private Context context;
    private BufferedReader reader;
    private PrintWriter writer;
    private boolean isRunning = false;

    /**
     * UnicastThread Constructor
     *
     * @param socket Unicast Socket
     */
    public UnicastThread(Socket socket) {
        this(socket, null);
    }

    /**
     * UnicastThread Constructor
     *
     * @param socket Unicast Socket
     * @param context Connection context
     */
    public UnicastThread(Socket socket, Context context) {
        this.socket = socket;
        this.context = context;
        initSocket();
    }

    /**
     * Initialise the Socket's Reader and Writer.
     *
     * @see BufferedReader
     * @see PrintWriter
     * @since 1.0
     */
    private void initSocket() {
        // Start the thread
        try {
            this.reader = new BufferedReader(new InputStreamReader(
                    this.socket.getInputStream(),
                    StandardCharsets.UTF_8
            ));
            this.writer = new PrintWriter(new OutputStreamWriter(
                    this.socket.getOutputStream(),
                    StandardCharsets.UTF_8
            ), true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Read from the socket.
     *
     * @return Returns the line that was read
     */
    protected String readLine() throws IOException {
        return this.reader.readLine();
    }

    /**
     * Write to the socket.
     *
     * @param str Content to write.
     */
    protected void write(String str) {
        this.writer.write(str);
        this.writer.flush();
    }

    /**
     * Print to the socket.
     *
     * @param str Content to print.
     */
    protected void print(String str) {
        this.writer.print(str);
        this.writer.flush();
    }

    /**
     * Getter, allow to retrieve an inputstream of the socket.
     *
     * @return Socket inputstream.
     */
    protected InputStream getInputStream() throws IOException {
        return this.socket.getInputStream();
    }

    /**
     * Getter, allow to retrieve an outputstream of the socket.
     *
     * @return Socket outputstream.
     */
    protected OutputStream getOutputStream() throws IOException {
        return this.socket.getOutputStream();
    }

    /**
     * Close all connection stream and resources.
     */
    @Override
    public void close() throws IOException {
        this.reader.close();
        this.writer.close();
        this.socket.close();
        setRunning(false);
    }

    /**
     * Allow to verify if the thread is running.
     * @return True; yes, False; no.
     */
    protected boolean isRunning() {
        return this.isRunning;
    }

    /**
     * Setter, define the running status.
     * @param running True; is running, False; is not running.
     */
    protected void setRunning(boolean running) {
        this.isRunning = running;
    }

    /**
     * Retrieve connection context.
     * @return Context.
     */
    protected Context getContext() {
        return this.context;
    }

    /**
     * Set connection context.
     * @param context context.
     */
    protected void setContext(Context context) {
        this.context = context;
    }
}
