package lightcontainer.domains.server;

import lightcontainer.domains.client.StoreProcessor;
import lightcontainer.interfaces.MulticastSPR;
import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.protocol.rules.reader.HelloRule;
import lightcontainer.repository.FileFrontEnd;
import lightcontainer.utils.Log;
import lightcontainer.utils.NetChooser;

import java.io.IOException;
import java.net.*;

/**
 * StoreMulticastRunnable
 * <p>
 * Class listening to the announcement of new StoreBackEnd.
 * Allowing it to be used as a storage unit.
 *
 * @author Jérémi NIHART <j.nihart@student.helmo.be>
 * @version 1.0
 * @see Runnable
 * @since 1.0
 */
public class MulticastServerListener implements Runnable {
    // Variable
    private final String multicast_address;
    private final int multicast_port;
    private final String network_interface;
    private final FileFrontEnd ffe;
    private final MulticastSPR repository;
    private final ProtocolRepository protocolRep;

    private final byte[] buffer = new byte[256];
    private MulticastSocket listener;

    // Constructor
    public MulticastServerListener(FileFrontEnd ffe, MulticastSPR repository, ProtocolRepository protocolRep, String multicast_address, int multicast_port, String network_interface) {
        this.ffe = ffe;
        this.repository = repository;
        this.protocolRep = protocolRep;
        this.multicast_address = multicast_address;
        this.multicast_port = multicast_port;
        this.network_interface = network_interface;
        repository.setServerListener(this);
    }

    /**
     * Start Multicast listening on indicated port and IP group.
     *
     * @see MulticastSocket#receive(DatagramPacket)
     * @see DatagramPacket
     * @since 1.0
     */
    @Override
    public void run() {
        try {
            // Create a new listening socket
            this.listener = new MulticastSocket(this.multicast_port);

            this.selectInterface();

            // Create an identifier for the multicast group on the specified ip
            InetAddress listener_group = InetAddress.getByName(this.multicast_address);
            // Creation of a packet for the information received
            DatagramPacket packet = new DatagramPacket(this.buffer, this.buffer.length);
            // Add the listener to the multicast group
            this.listener.joinGroup(listener_group);
            while (true) {
                // Read the packet received and build a string of characters
                this.listener.receive(packet);

                onNewSbe(packet);
            }
        } catch (IOException ioException) {
            Log.getInstance().infoln("[ERREUR] Multicast server can't start : " + ioException.getMessage());
        }
    }

    private void onNewSbe(DatagramPacket packet) {
        try {
            String data = new String(packet.getData(), 0, packet.getLength());

            HelloRule.Result readerResult = protocolRep.executeReader(null, data);

            Log.getInstance().infof("Nouveau SBE : Domain=%s | Port=%d\n", readerResult.getDomain(), readerResult.getPort());

            if (!this.repository.hasDomain(readerResult.getDomain())){
                Socket socket = new Socket(packet.getAddress(), readerResult.getPort());

                // Create the store processor
                StoreProcessor storeProcessor = new StoreProcessor(socket, readerResult.getDomain(), ffe, protocolRep); // TODO <!!!> : Voir comment on procède get via repo ou ici ?!

                // Add the store processor to its repository
                this.repository.addStore(storeProcessor);
            }
            // Contient déjà le SBE donc maj de la dernière activité
            this.repository.updateLastAnnounce(readerResult.getDomain());
        } catch (IOException | ClassCastException exception) {
            Log.getInstance().infoln("[ERREUR] Une SBE essaye de se connecter avec une mauvaise configuration : " + exception.getMessage());
        }
    }

    /**
     * Permet de choisir l'interface réseau d'écoute en Multicast
     */
    private void selectInterface() {
        try {
            this.listener.setNetworkInterface(NetworkInterface.getByName(network_interface));
        } catch (SocketException| NullPointerException e) {
            NetChooser netChooser = new NetChooser();
            try {
                this.listener.setNetworkInterface(netChooser.requestInterface());
            } catch (SocketException socketException) {
                socketException.printStackTrace();
                this.stop();
            }
        }
    }

    /**
     * Closes the MulticastSocket connection and aborts the listening and infinite listening loop.
     *
     * @see MulticastServerListener#run()
     * @since 1.0
     */
    public void stop() {
        repository.disconnectDomains();
        this.listener.close();
    }
}
