package lightcontainer.repository;

import lightcontainer.domains.Task;
import lightcontainer.domains.client.StoreProcessor;
import lightcontainer.domains.server.MulticastServerListener;
import lightcontainer.interfaces.MulticastSPR;

import java.time.LocalDateTime;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
// TODO : C'est genre un ClientHandlerManager quoi hein, normal qu'il fasse blinder de chose ;)

/**
 * StoreProcessorRepository
 * <p>
 * Repository storing StorePorcessor class.
 * Contains some utility functions.
 *
 * @author Jérémi NIHART <j.nihart@student.helmo.be>
 * @version 1.0
 * @see StoreProcessor
 * @since 1.0
 */
public class StoreProcessorRepository implements AutoCloseable, MulticastSPR {
    // Variables
    private final Set<StoreProcessor> handlers;
    private MulticastServerListener server;

    // Constructor
    public StoreProcessorRepository() {
        this.handlers = new HashSet<>();
    }

    /**
     * Setter, allow to define the ServerListener of a repository.
     * & start the server.
     *
     * @param server ServerListener to set as default.
     * @since 1.0
     */
    @Override
    public void setServerListener(MulticastServerListener server) {
        this.server = server;
        new Thread(server).start();
    }

    /**
     * Add a StorePorcessor.
     *
     * @param store Store processor to add.
     * @since 1.0
     */
    @Override
    public boolean addStore(StoreProcessor store) {
        if (!this.hasDomain(store.getDomain())) {
            store.start();
            return this.handlers.add(store);
        }
        return false;
    }

    @Override
    public String findDomain() {
        StoreProcessor handler = findSBE();
        return handler == null ? null : handler.getDomain();
    }

    private StoreProcessor findSBE() {
        for (StoreProcessor handler : handlers) {
            if (handler.canProcessTask()) {
                return handler;
            }
        }

        return null;
    }

    private StoreProcessor getSBE(String stor) {
        for (StoreProcessor handler : handlers) {
            if (handler.getDomain().equals(stor)) {
                return handler;
            }
        }

        return null;
    }

    @Override
    public void assignTask(String stor, Task task) {
        StoreProcessor handler = getSBE(stor);
        if (handler != null) {
            handler.executeCommand(task.getContext(), task.getCommand());
        }
    }

    @Override
    public boolean hasDomain(String domain) {
        for (StoreProcessor handler : handlers) {
            if (handler.getDomain().equals(domain)) {
                return true;
            }
        }

        return false;
    }

    @Override
    public int domainCount() {
        return handlers.size();
    }

    @Override
    public void disconnectDomains() {
        for (StoreProcessor handler : handlers) {
            handler.close();
        }
    }

    /**
     * Permet de déconnecter un SBE
     *
     * @param domain Le domaine du SBE à déconnecter
     */
    @Override
    public void closeStore(String domain) {
        StoreProcessor storeProcessor = getSBE(domain);
        if (storeProcessor != null) {
            storeProcessor.close();
            handlers.remove(storeProcessor);
        }
    }

    /**
     * Permet de mettre à jours la dernière annonce de ce SBE au FFE
     *
     * @param domain Le domain s'annoncant
     */
    @Override
    public void updateLastAnnounce(String domain) {
        StoreProcessor storeProcessor = getSBE(domain);
        if (storeProcessor != null) {
            storeProcessor.setLastAnnounce(LocalDateTime.now());
        }
    }

    /**
     * Permet de récupérer les noms des domaines connectés au FFE
     *
     * @return Les noms des domaines connectés au FFE
     */
    @Override
    public Collection<String> getDomains() {
        Set<String> domains = new HashSet<>();

        for (StoreProcessor domain : handlers) {
            domains.add(domain.getDomain());
        }

        return domains;
    }

    /**
     * Permet de récupérer la dernière annonce d'un SBE
     *
     * @param domain
     * @return La dernière annonce d'un SBE
     */
    @Override
    public LocalDateTime getLastAnnounce(String domain) {
        StoreProcessor storeProcessor = getSBE(domain);
        return storeProcessor == null ? null : storeProcessor.getLastAnnounce();
    }

    /**
     * AutoClosable Function
     * Closes all StoreProcessor stored in this repository and deallocates all resources.
     *
     * @since 1.0
     */
    @Override
    public void close() {
        // Stop the server.
        this.server.stop();
        // Close each client.
        this.handlers.forEach(StoreProcessor::close);
    }



}
