package lightcontainer.storage;

import com.google.gson.*;
import lightcontainer.utils.Log;

import java.util.*;

/**
 * Specific implementation of Adapter that converts AppData to Json and vice-versa
 */
public class JsonAdapter implements Adapter {

    /**
     * @return A Json String containing AppData properties
     */
    @Override
    public String toString(AppData appData) {
        return addData(appData);
    }

    private String addData(AppData appData) {
        AppConfig appConfig = appData.getAppConfig();
        JsonObject config = new JsonObject();
        config.addProperty("unicast_port", appConfig.getUnicastPort());
        config.addProperty("multicast_ip", appConfig.getMulticastIp());
        config.addProperty("multicast_port", appConfig.getMulticastPort());
        config.addProperty("network_interface", appConfig.getNetworkInterface());
        config.addProperty("tls", appConfig.isTls());
        config.addProperty("storagePath", appConfig.getStoragePath());
        JsonArray users = new JsonArray();
        Iterator<User> userIterator = appData.usersIterator();
        addUsers(users, userIterator);
        config.add("users", users);
        return config.toString();
    }

    private void addUsers(JsonArray users, Iterator<User> userIterator) {
        while (userIterator.hasNext()) {
            User current = userIterator.next();
            JsonObject user = new JsonObject();
            user.addProperty("name", current.getName());
            user.addProperty("password", current.getPassword());
            user.addProperty("aes_key", current.getAesKey());
            JsonArray files = new JsonArray();
            Iterator<File> fileIterator = current.fileIterator();
            addFiles(fileIterator, files);
            user.add("files", files);
            users.add(user);
        }
    }

    private void addFiles(Iterator<File> fileIterator, JsonArray files) {
        while (fileIterator.hasNext()) {
            File currentFile = fileIterator.next();
            JsonObject file = new JsonObject();
            file.addProperty("name", currentFile.getName());
            file.addProperty("fileNameSalt", currentFile.getFileNameSalt());
            file.addProperty("size", currentFile.getSize());
            file.addProperty("iv", currentFile.getIv());
            JsonArray storage = new JsonArray();
            Iterator<String> storageIterator = currentFile.getStorageIterator();
            addStorage(storage, storageIterator);
            file.add("storage", storage);
            files.add(file);
        }
    }

    private void addStorage(JsonArray storage, Iterator<String> storageIterator) {
        while (storageIterator.hasNext()) {
            String storageString = storageIterator.next();
            storage.add(storageString);
        }
    }

    /**
     * @param appDataString The Json String to convert
     * @return An AppData instance
     */
    @Override
    public AppData fromString(String appDataString) {
        try {
            JsonElement jsonString = JsonParser.parseString(appDataString);
            JsonObject jsonAppData = jsonString.getAsJsonObject();
            AppConfig appConfig = AppConfig.getInstance();
            appConfig.setUnicastPort(jsonAppData.get("unicast_port").getAsInt());
            appConfig.setMulticastIp(jsonAppData.get("multicast_ip").getAsString());
            appConfig.setMulticastPort(jsonAppData.get("multicast_port").getAsInt());
            appConfig.setNetworkInterface(jsonAppData.get("network_interface").getAsString());
            appConfig.setTls(jsonAppData.get("tls").getAsBoolean());
            appConfig.setStoragePath(jsonAppData.get("storagePath").getAsString());
            JsonArray jsonUsers = jsonAppData.getAsJsonArray("users");
            List<User> users = new ArrayList<>();
            getUsers(jsonUsers, users);
            AppData appData = AppData.getInstance();
            appData.setAppConfig(appConfig);
            for (User user : users) {
                appData.addUser(user);
            }
            return appData;
        } catch (JsonParseException parseException) {
            Log.getInstance().infoln("[FFE] : Error while loading configuration file"); //TODO - changer en log
            return null;
        }
    }

    private void getUsers(JsonArray jsonUsers, List<User> users) {
        for (JsonElement element : jsonUsers) {
            JsonObject jsonUser = element.getAsJsonObject();
            String name = jsonUser.get("name").getAsString();
            String password = jsonUser.get("password").getAsString();
            String aeskey = jsonUser.get("aes_key").getAsString();
            Map<String, File> userFiles = new HashMap<>();
            JsonArray jsonFiles = jsonUser.getAsJsonArray("files");
            getFiles(userFiles, jsonFiles);
            User user = new User(name, password, aeskey, userFiles);
            users.add(user);
        }
    }

    private void getFiles(Map<String, File> userFiles, JsonArray jsonFiles) {
        for (JsonElement fileElement : jsonFiles) {
            JsonObject jsonFile = fileElement.getAsJsonObject();
            String fileName = jsonFile.get("name").getAsString();
            String fileNameSalt = jsonFile.get("fileNameSalt").getAsString();
            int size = jsonFile.get("size").getAsInt();
            String iv = jsonFile.get("iv").getAsString();
            Set<String> storage = new HashSet<>();
            JsonArray jsonStorage = jsonFile.getAsJsonArray("storage");
            getStorage(storage, jsonStorage);
            File file = new File(fileName, fileNameSalt, size, iv, storage);
            userFiles.put(file.getName(), file);
        }
    }

    private void getStorage(Set<String> storage, JsonArray jsonStorage) {
        for (JsonElement storageElement : jsonStorage) {
            String storageName = storageElement.getAsString();
            storage.add(storageName);
        }
    }
}
