package lightcontainer.utils;

import javax.crypto.*;
import javax.crypto.spec.GCMParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.security.*;
import java.util.Base64;

/**
 * AES GCM 256 Encryption Class [DO NOT EDIT]
 *
 * @author Jérémi Nihart <contact@endmove.eu>
 * @version 1.0
 * @since 1.0
 */
public class AES_GCM {
    // Constants
    public static final int AES_KEY_SIZE = 256;
    public static final int GCM_IV_LENGTH = 16;
    public static final int GCM_TAG_LENGTH = 16;


    /**
     * Decoder to decode base64 vector to byte vector.
     *
     * @param base64Vector A base64 encoded vector.
     * @return Byte vector.
     */
    private static byte[] decodeBase64Vector(String base64Vector) {
        Base64.Decoder b64Decoder = Base64.getDecoder();
        return b64Decoder.decode(base64Vector);
    }

    /**
     * Decoder to decode base64 string to plain string.
     *
     * @param base64String A base64 encoded string.
     * @return Plain string.
     * @see AES_GCM#decodeBase64Vector(String)
     */
    private static String decodeBase64String(String base64String) {
        return new String(decodeBase64Vector(base64String));
    }

    /**
     * Encoder to encode vector to base64 string.
     *
     * @param rawVector A raw vector.
     * @return A base64 encoded vector.
     */
    private static String encodeBase64(byte[] rawVector) {
        Base64.Encoder b64Encoder = Base64.getEncoder();
        return b64Encoder.encodeToString(rawVector);
    }

    /**
     * Encoder to encode string to base64 string.
     *
     * @param rawString A raw string.
     * @return A base64 encoded string.
     * @see AES_GCM#encodeBase64(byte[]))
     */
    private static String encodeBase64(String rawString) {
        return encodeBase64(rawString.getBytes(StandardCharsets.UTF_8));
    }

    /**
     * FACTORY, to setting up a Java cryptographic cypher.
     *
     * @param op_mode the operation mode of this cipher (this is one of the
     *                following: ENCRYPT_MODE, DECRYPT_MODE, WRAP_MODE or UNWRAP_MODE)
     * @param key     Base64 encoded secret key.
     * @param IV      Base64 encoded vector.
     * @return A Cryptography cypher.
     * @throws AesGcmException Throw an exception in case of an error occur while setting up the the cypher.
     */
    private static Cipher createCipher(int op_mode, String key, String IV) throws AesGcmException {
        try {
            // Get a cipher instance
            Cipher cipher = Cipher.getInstance("AES/GCM/NoPadding");
            // Create a secret key from the key for the specific crypto algo
            SecretKeySpec keySpec = new SecretKeySpec(decodeBase64Vector(key), "AES");
            // Create a GCMParameterSpec to setting up the AES init vector
            GCMParameterSpec gcmParameterSpec = new GCMParameterSpec(GCM_TAG_LENGTH * 8, decodeBase64Vector(IV));
            // Init the cipher for the selected operation
            cipher.init(op_mode, keySpec, gcmParameterSpec);
            return cipher;
        } catch (InvalidAlgorithmParameterException | NoSuchPaddingException | NoSuchAlgorithmException | InvalidKeyException e) {
            throw new AesGcmException(e, "Error while creating the Cipher object");
        }
    }

    /**
     * Generate a secret key base64 encoded.
     *
     * @return New Secret key b64 encoded.
     * @throws AesGcmException Exception if an error occur.
     */
    public static String generateSecretKey() throws AesGcmException {
        try {
            KeyGenerator keyGenerator = KeyGenerator.getInstance("AES");
            keyGenerator.init(AES_KEY_SIZE);
            SecretKey key = keyGenerator.generateKey();
            return encodeBase64(key.getEncoded());
        } catch (NoSuchAlgorithmException e) {
            throw new AesGcmException("Error while generating the AES secret key");
        }
    }

    /**
     * Generate an IV (initialisation vector) base64 encoded.
     *
     * @return New generated IV b64 encoded.
     */
    public static String generateIV() {
        byte[] IV = new byte[GCM_IV_LENGTH];
        SecureRandom random = new SecureRandom();
        random.nextBytes(IV);
        return encodeBase64(IV);
    }

    /**
     * Encrypt text, with AES GCM.
     *
     * @param plainText Plain text to encrypt.
     * @param key       Base64 encoded secret key.
     * @param IV        Base64 encoded vector.
     * @return The encrypted plain text Base64 encoded.
     * @throws AesGcmException Exception if an error occur.
     */
    public static String encrypt(String plainText, String key, String IV) throws AesGcmException {
        try {
            // Make the cipher for encryption
            Cipher cipher = createCipher(Cipher.ENCRYPT_MODE, key, IV);
            // Perform Encryption
            return encodeBase64(cipher.doFinal(plainText.getBytes()));
        } catch (Exception e) {
            throw new AesGcmException(e);
        }
    }

    /**
     * Encrypt stream, with AES GCM.
     * The output stream is automatically closed after processing.
     *
     * @param in       InputStream to the input, flux to encrypt.
     * @param out      OutputStream to the output, encrypted flux.
     * @param fileSize Stream/file size (! unencrypted size !).
     * @param key      Base64 encoded secret key.
     * @param IV       Base64 encoded vector.
     * @throws AesGcmException Exception if an error occur.
     */
    public static void encryptStream(InputStream in, OutputStream out, long fileSize, String key, String IV) throws AesGcmException {
        byte[] buffer = new byte[1024];
        long currentSize = 0;
        int bytes;
        try {
            // Make the cipher for encryption
            Cipher cipher = createCipher(Cipher.ENCRYPT_MODE, key, IV);
            // Initialize a CipherOutputStream
            CipherOutputStream cipherOut = new CipherOutputStream(out, cipher);
            // Encryption Process
            while ((currentSize < fileSize) && (bytes = in.read(buffer)) > 0) {
                cipherOut.write(buffer, 0, bytes);
                cipherOut.flush();
                currentSize += bytes;
            }
            // Close CipherOutputStream
            cipherOut.close();
        } catch (Exception e) {
            throw new AesGcmException(e);
        }
    }

    /**
     * Decrypt, with AES GCM.
     *
     * @param cryptText The encrypted text.
     * @param key       Base64 encoded secret key.
     * @param IV        Base64 encoded vector.
     * @return The decrypted plain text.
     * @throws AesGcmException Exception if an error occur.
     */
    public static String decrypt(String cryptText, String key, String IV) throws AesGcmException {
        try {
            // Make the cipher for decryption
            Cipher cipher = createCipher(Cipher.DECRYPT_MODE, key, IV);
            // Perform Decryption
            return new String(cipher.doFinal(decodeBase64Vector(cryptText)));
        } catch (Exception e) {
            throw new AesGcmException(e);
        }
    }

    /**
     * Decrypt stream, with AES GCM.
     * The input stream is automatically closed after processing.
     *
     * @param in       InputStream to the input, flux to decrypt.
     * @param out      OutputStream to the output, decrypted flux.
     * @param fileSize Stream/file size (! encrypted size !).
     * @param key      Base64 encoded secret key.
     * @param IV       Base64 encoded vector.
     * @throws AesGcmException Exception if an error occur.
     */
    public static void decryptStream(InputStream in, OutputStream out, long fileSize, String key, String IV) throws AesGcmException {
        byte[] buffer = new byte[1024];
        long currentSize = 0;
        int bytes;
        try {
            // Make the cipher for decryption
            Cipher cipher = createCipher(Cipher.DECRYPT_MODE, key, IV);
            // Initialize a CipherOutputStream
            CipherInputStream cipherIn = new CipherInputStream(in, cipher);
            // Encryption Process
            while ((currentSize < fileSize) && (bytes = cipherIn.read(buffer)) > 0) {
                out.write(buffer, 0, bytes);
                out.flush();
                currentSize += bytes;
            }
            // Close CipherOutputStream
            cipherIn.close();
        } catch (Exception e) {
            throw new AesGcmException(e);
        }
    }

    /**
     * Internal Error from AES_GCM encryption Class
     */
    public static class AesGcmException extends Exception {
        // Constant
        private static final long serialVersionUID = -145972354893514657L;

        /**
         * Constructor of AesGcmException,
         * which define it's own detail message.
         *
         * @param msg The detail message.
         */
        public AesGcmException(String msg) {
            super(msg);
        }

        /**
         * Constructor of AesGcmException,
         * which propagates the error triggering
         * a crash of the encryption system.
         *
         * @param e Previous exception throwable.
         */
        public AesGcmException(Throwable e) {
            super(e);
        }

        /**
         * Constructor of AesGcmException,
         * which propagates the error triggering
         * a crash of the encryption system with
         * a chosen detail message.
         *
         * @param e   Previous exception throwable.
         * @param msg The detail message.
         */
        public AesGcmException(GeneralSecurityException e, String msg) {
            super(msg, e);
        }
    }
}