package lightcontainer.utils;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.Arrays;

/**
 * DeepFileEraser - Secure file remover
 *
 * @since 1.0
 * @version 1.0
 *
 * @author Jérémi Nihart <contact@endmove.eu>
 */
public class DeepFileEraser {
    public static final int DFE_BUFFER_DEFAULT_SIZE = 1024;         // Default buffer size
    public static final int DFE_BUFFER_MINIMUM_SIZE = 250;          // Minimum buffer size
    public static final int DFE_BASE_DEFAULT_FILL_VALUE = 0;        // Filling value of the initialisation phase
    public static final int DFE_STEP_MIN_FILL_VALUE = 0;            // Minimum filling value of a writing phase
    public static final int DFE_STEP_MAX_FILL_VALUE = 15;           // Maximum filling value of a writing phase


    /**
     * Generate a number randomly.
     *
     * @param           min Minimal value (included)
     * @param           max Maximal value (included)
     *
     * @return          Random generated number.
     */
    private static int random(int min, int max) {
        return min + (int)(Math.random() * ((max - min) + 1));
    }

    /**
     * Simple remove operation.
     *
     * @param               target File object on which to operate.
     * @param               length File size.
     *
     * @throws              DeepFileEraserException Exception if an error occur.
     */
    private static void eraseBase(File target, long length) throws DeepFileEraserException {
        byte[] buffer = new byte[DFE_BUFFER_DEFAULT_SIZE];
        long wroteLength = 0;
        try (OutputStream out = new FileOutputStream(target)) {
            Arrays.fill(buffer, (byte)DFE_BASE_DEFAULT_FILL_VALUE);
            while (wroteLength < length) {
                if (wroteLength + buffer.length > length) {
                    buffer = new byte[(int)(length-wroteLength)];
                    Arrays.fill(buffer, (byte)DFE_BASE_DEFAULT_FILL_VALUE);
                }
                out.write(buffer);
                wroteLength += buffer.length;
            }
        } catch (Exception e) {
            throw new DeepFileEraserException(e, "Error while the Base erasing step.");
        }
    }

    /**
     * Deep remove operation.
     *
     * @param               target File object on which to operate.
     * @param               length File size.
     *
     * @throws              DeepFileEraserException Exception if an error occur.
     */
    private static void eraseStep(File target, long length) throws DeepFileEraserException {
        byte[] buffer = new byte[random(DFE_BUFFER_MINIMUM_SIZE, DFE_BUFFER_DEFAULT_SIZE)];
        long wroteLength = 0;
        try (OutputStream out = new FileOutputStream(target)) {
            while(wroteLength < length) {
                if (wroteLength + buffer.length > length) buffer = new byte[(int)(length-wroteLength)];
                Arrays.fill(buffer, (byte) random(DFE_STEP_MIN_FILL_VALUE, DFE_STEP_MAX_FILL_VALUE));
                out.write(buffer);
                wroteLength += buffer.length;
            }
        } catch (Exception e) {
            throw new DeepFileEraserException(e, "Error while the Step erasing step.");
        }
    }

    /**
     * Erase a file deeply.
     * default deeply level = 7.
     *
     * @param               filePath Path of the file.
     *
     * @throws              DeepFileEraserException Exception if an error occur.
     */
    public static void eraseFile(String filePath) throws DeepFileEraserException {
        eraseFile(filePath, 7);
    }

    /**
     * Erase a file deeply.
     *
     * @param               filePath Path of the file.
     * @param               depthLevel Erasure level (the higher the number, the longer and more secure the erasure).
     *
     * @throws              DeepFileEraserException Exception if an error occur.
     */
    public static void eraseFile(String filePath, int depthLevel) throws DeepFileEraserException {
        // Init
        File target = new File(filePath);
        if (target.exists()) {
            long targetSize = target.length();
            // Base write
            eraseBase(target, targetSize);
            // Deep write
            for (int i = 0; i < depthLevel; i++) eraseStep(target, targetSize);
            // Delete file
            target.delete();
        }
    }

    /**
     * Internal Error from DeepFileEraser encryption Class
     */
    public static class DeepFileEraserException extends Exception {
        // Constant
        private static final long serialVersionUID = -158979547897543145L;

        /**
         * Constructor of DeepFileEraserException,
         * which propagates the error triggering
         * a crash of the erasing file system with
         * a chosen detail message.
         *
         * @param           e Previous exception throwable.
         * @param           msg The detail message.
         */
        public DeepFileEraserException(Throwable e, String msg) {
            super(msg, e);
        }
    }
}
